#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
""" Tests for the product fetcher """

# pylint: disable=C0103, E0401, E0402, E1136, R0201, R0903

from pathlib import Path
from typing import List, Union

import pytest
from productfetcher.fetchers import DryRunFakeFileFetcher, ForceMode
from productfetcher.interfaces import (
    FetcherFactory,
    FetchPlan,
    FileFetcher,
    LocatedFile,
)
from productfetcher.locations import FileLocator, ServiceLocator
from productfetcher.product_fetcher import CLIParam, FetchContext


class FakePlan(FetchPlan):
    """A non-fetching fetch plan"""

    def __init__(self, fetchers, concurrency):
        self.fetchers = fetchers
        self.concurrency = concurrency

    def fetch(self):
        # Let's not actually fetch anything today
        pass


class FakeFactory(FetcherFactory):
    """Creates non-fetching fetchers"""

    def fetch_ngas(self, file: LocatedFile) -> FileFetcher:
        return DryRunFakeFileFetcher(file)

    def fetch_oracle_xml(self, file: LocatedFile) -> FileFetcher:
        return DryRunFakeFileFetcher(file)

    def fetch_plan(self, fetchers: List[Union[FileFetcher, FetchPlan]], concurrency: int = 1) -> FetchPlan:
        return FakePlan(fetchers, concurrency)


def test_plan_generation(resource_path_root):
    """
    Can we generate a test plan?

    :param resource_path_root:
    :return:
    """
    report = FileLocator(resource_path_root / "location_files" / "13B-014.json").locate()
    fake_factory = FakeFactory()
    fetchers = report.fetchers(fake_factory)

    # we should absolutely have the same number of fetchers as we have files
    assert len(fetchers) == len(report.files)

    # let's make a plan
    plan = FetchContext.calculate_plan(fake_factory, fetchers)

    # we should have the same number of fetchers in the plan as in the original report
    assert sum(len(subplan.fetchers) for subplan in plan.fetchers) == len(report.files)


def test_multiple_locator_fetching(capsys, resource_path_root):
    """
    Can we handle multiple locators?

    :param capsys:
    :return:
    """
    img = resource_path_root / "location_files" / "IMG.json"
    cal = resource_path_root / "location_files" / "CALIBRATION.json"

    # parse the command line with these two
    fc = FetchContext.parse_commandline([CLIParam.FILE.value, str(img), CLIParam.FILE.value, str(cal)])
    assert len(fc.locators) == 2

    # let's make the plan and ensure we have all the stuff we expect from both
    plan = fc.generate_plan()

    # we'll need to open these files ourselves to figure out what fetchers we expect
    for locator_file in [img, cal]:
        for file in FileLocator(locator_file).locate().files:
            seen = False

            # there may be a more "test friendly" way of doing this, such as by asking the plan
            # if it is fetching a certain file, but it seems like a lot of refactoring for this
            # one test, so I'm going to leave it alone for now
            for hostgroup_fetcher in plan.fetchers:
                seen = seen or file in [fetcher.file for fetcher in hostgroup_fetcher.fetchers]
            assert seen


def test_argument_parsing(capsys):
    """
    Can we parse the command-line arguments passed in?

    :param capsys:
    :return:
    """
    # ensure that various combinations do not work
    with pytest.raises(SystemExit):
        # must have an SPL or a file
        FetchContext.parse_commandline([])

    # check the dry run value
    fc = FetchContext.parse_commandline([CLIParam.DRY.value, CLIParam.FILE.value, "foo"])
    assert fc.dry_run
    fc = FetchContext.parse_commandline([CLIParam.FILE.value, "foo"])
    assert not fc.dry_run

    # check the force value
    fc = FetchContext.parse_commandline([CLIParam.FORCE.value, CLIParam.FILE.value, "foo"])
    assert fc.force == ForceMode.FORCE
    fc = FetchContext.parse_commandline([CLIParam.FILE.value, "foo"])
    assert fc.force == ForceMode.NORMAL

    fc = FetchContext.parse_commandline([CLIParam.FILE.value, "foo"])
    assert isinstance(fc.locators[0], FileLocator)
    assert fc.locators[0].file == Path("foo")

    fc = FetchContext.parse_commandline([CLIParam.SPL.value, "uid://this/is/a/fakesy"])
    assert isinstance(fc.locators[0], ServiceLocator)
    assert fc.locators[0].spl == "uid://this/is/a/fakesy"

    fc = FetchContext.parse_commandline([CLIParam.FILE.value, "foo", CLIParam.CONCURRENCY.value, "732"])
    assert fc.concurrency == 732

    capsys.readouterr()
